/******************************************************************************
 * Main.cpp
 ******************************************************************************
 $Header:: /N-Patch/GlutViewer/Main.cpp 4     4/13/01 12:33a Avlachos         $
 ******************************************************************************
 * Created by Alex Vlachos (AVlachos@ati.com) on 4-12-2001
 * (C) 2000 ATI Research, Inc.  All rights reserved.
 ******************************************************************************/

static char sgCopyrightString[] = "\r\n\r\n(C) 2000 ATI Research, Inc.\r\n\r\n";

// INCLUDES ===================================================================
#include <windows.h>
#include <winuser.h>
#include <process.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\glut.h>
#include "types.h"

// STRUCTURES =================================================================
#pragma pack (push)
#pragma pack (1) //Don't pad the following structures

typedef union
{
   //Point32 in 3D space
   struct { float32 x, y, z; };
   struct { float32 v[3]; };
} AtiVector3, Vec3;

typedef union
{
   //Point32 in 2D space
   struct { float32 u, v; };
   struct { float32 t[2]; };
} AtiTexCoord2D;

typedef struct //Preprocess polygon
{
   AtiVector3 vertex[3];
   AtiVector3 normal[3];
   AtiTexCoord2D uv[3];
} NPatchPolygon;

typedef struct //Preprocess polygon
{
   AtiVector3 vertex;
   AtiVector3 normal;
   AtiTexCoord2D uv;
} NPatchVertex;

#pragma pack (pop)

// MACROS =====================================================================
#define AtiMallocTypeArray(array,num,type)                               \
   array = NULL;                                                         \
   if ((num) > 0)                                                        \
   {                                                                     \
      array = new type [num];                                            \
      if (array == NULL)                                                 \
      {                                                                  \
         printf("\nERROR: Can't allocate memory (%d bytes) for " #array "!\n", sizeof(type) * (num)); \
         return;                                                         \
      }                                                                  \
      memset (array, 0, sizeof(type) * (num));                           \
   }

/*array = (type *) malloc (sizeof(type) * (num));*/
/*free(array);*/

#define AtiFreeTypeArray(array) \
   if (array != NULL)           \
      delete [] array;          \
   array = NULL;

// GLOBALS ====================================================================
char8 gCurrentFile[64] = "1.m";

bool8 gDrawNormals = FALSE;
bool8 gDrawWireframe = FALSE;
bool8 gLinearNormals = TRUE;
bool8 gDrawSpecular = TRUE;
bool8 gAutoLODMode = FALSE;
bool8 gShowHelp = FALSE;
int32 gLevel = 10;

int32 gWindowWidth = 640;
int32 gWindowHeight = 480;

RECT gDesktopRect;

//Camera position
Vec3 gCamera;
float32 gYaw = 0.0f;
float32 gPitch = 15.0f;

//Geometry
int32 gNumVertices = 0;
int32 gNumVerticesAllocated = 0;
NPatchVertex *gVertexArray = NULL;

int32 gNumIndices = 0;
int32 gNumIndicesAllocated = 0;
uint16 *gIndexArray = NULL;

int32 gNumRenderVertices = 0;
NPatchVertex *gRenderVertexArray = NULL;

int32 gNumRenderIndices = 0;
uint16 *gRenderIndexArray = NULL;

float32 gZoomDistance = 25.0f;
float32 gPanX = 0.0f;
float32 gPanY = 0.0f;

// FUNCTIONS ==================================================================
void EvaluateBezierPoint(float32 coeff[][3], float32 normCoeff[][3], float32 b1, float32 b2, float32 b3, AtiVector3 *destVertices, AtiVector3 *destNormals)
{
   int32 i;
   int32 x, y;
   float32 buff[6][3];
   float32 tmpf;

   x = 0;
   y = 1;
   for (i=0; i<6; i++)
   {
      buff[i][0] = coeff[x][0]*b2 + coeff[y][0]*b1 + coeff[y+1][0]*b3;
      buff[i][1] = coeff[x][1]*b2 + coeff[y][1]*b1 + coeff[y+1][1]*b3;
      buff[i][2] = coeff[x][2]*b2 + coeff[y][2]*b1 + coeff[y+1][2]*b3;

      x++;
      if ((i == 0) || (i == 2))
         y += 2;
      else
         y++;
   }

   x = 0;
   y = 1;
   for (i=0; i<3; i++)
   {
      buff[i][0] = buff[x][0]*b2 + buff[y][0]*b1 + buff[y+1][0]*b3;
      buff[i][1] = buff[x][1]*b2 + buff[y][1]*b1 + buff[y+1][1]*b3;
      buff[i][2] = buff[x][2]*b2 + buff[y][2]*b1 + buff[y+1][2]*b3;

      x++;
      if (i == 0)
         y += 2;
      else
         y++;
   }

   destVertices->x = buff[0][0]*b2 + buff[1][0]*b1 + buff[2][0]*b3;
   destVertices->y = buff[0][1]*b2 + buff[1][1]*b1 + buff[2][1]*b3;
   destVertices->z = buff[0][2]*b2 + buff[1][2]*b1 + buff[2][2]*b3;

   /**********/
   /* Normal */
   /**********/
   x = 0;
   y = 1;
   for (i=0; i<3; i++)
   {
      buff[i][0] = normCoeff[x][0]*b2 + normCoeff[y][0]*b1 + normCoeff[y+1][0]*b3;
      buff[i][1] = normCoeff[x][1]*b2 + normCoeff[y][1]*b1 + normCoeff[y+1][1]*b3;
      buff[i][2] = normCoeff[x][2]*b2 + normCoeff[y][2]*b1 + normCoeff[y+1][2]*b3;

      x++;
      if (i == 0)
         y += 2;
      else
         y++;
   }

   destNormals->x = buff[0][0]*b2 + buff[1][0]*b1 + buff[2][0]*b3;
   destNormals->y = buff[0][1]*b2 + buff[1][1]*b1 + buff[2][1]*b3;
   destNormals->z = buff[0][2]*b2 + buff[1][2]*b1 + buff[2][2]*b3;
   tmpf = 1.0f / sqrt(destNormals->x*destNormals->x + destNormals->y*destNormals->y + destNormals->z*destNormals->z);
   destNormals->x *= tmpf;
   destNormals->y *= tmpf;
   destNormals->z *= tmpf;
}

//=============================================================================
void CreateNPatchGeometry (int32 level)
{
   int32 i;
   int32 j;
   int32 k;

   if (level < 0)
      level = 0;

   /***************/
   /* Free Memory */
   /***************/
   AtiFreeTypeArray(gRenderVertexArray);
   gNumRenderVertices = 0;

   AtiFreeTypeArray(gRenderIndexArray);
   gNumRenderIndices = 0;

   /***********************/
   /* Allocate new memory */
   /***********************/
   int32 numNewPolygons = 0;
   int32 numNewVertices = 65535;
   while (numNewVertices >= 65535)
   {
      numNewPolygons = 0;
      for (i=0; i<=level; i++)
         numNewPolygons += (i*2)+1;
      numNewPolygons *= gNumIndices/3;
      //printf("%d\n", numNewPolygons);

      numNewVertices = 3;
      for (i=0; i<level; i++)
         numNewVertices += i+3;
      numNewVertices *= gNumIndices/3;
      //printf("%d\n", numNewVertices);

      if (level <= 0)
         break;
      if (numNewVertices >= 65535)
         level--;
   }
   gLevel = level;

   AtiMallocTypeArray(gRenderVertexArray, numNewVertices, NPatchVertex);
   AtiMallocTypeArray(gRenderIndexArray, numNewPolygons*3, uint16);

   /*************************************************/
   /* Create new geometry for each original polygon */
   /*************************************************/
   AtiVector3 *destVertices = &(gRenderVertexArray[0].vertex);
   AtiVector3 *destNormals = &(gRenderVertexArray[0].normal);
   AtiTexCoord2D *destUVs = &(gRenderVertexArray[0].uv);
   uint16 *destIndices = gRenderIndexArray;

   for (i=0; i<gNumIndices/3; i++) //For each original polygon
   {
      float32 tmpf;
      float32 coeff[10][3];
      float32 normCoeff[6][3];

      AtiVector3 *v1 = &(gVertexArray[gIndexArray[i*3+0]].vertex);
      AtiVector3 *v2 = &(gVertexArray[gIndexArray[i*3+1]].vertex);
      AtiVector3 *v3 = &(gVertexArray[gIndexArray[i*3+2]].vertex);

      AtiVector3 *n1 = &(gVertexArray[gIndexArray[i*3+0]].normal);
      AtiVector3 *n2 = &(gVertexArray[gIndexArray[i*3+1]].normal);
      AtiVector3 *n3 = &(gVertexArray[gIndexArray[i*3+2]].normal);

      AtiTexCoord2D *uv1 = &(gVertexArray[gIndexArray[i*3+0]].uv);
      AtiTexCoord2D *uv2 = &(gVertexArray[gIndexArray[i*3+1]].uv);
      AtiTexCoord2D *uv3 = &(gVertexArray[gIndexArray[i*3+2]].uv);

      AtiVector3 edge;
      AtiVector3 avgNorm;
      AtiVector3 tangent1;
      AtiVector3 tangent2;

      /***************************************************/
      /* Compute cubic Bezier control mesh for positions */
      /***************************************************/
      //    v2  0
      //       1 2
      //      3 4 5
      // v1  6 7 8 9  v3

      //3 corner vertices
      coeff[6][0] = v1->x;
      coeff[6][1] = v1->y;
      coeff[6][2] = v1->z;

      coeff[0][0] = v2->x;
      coeff[0][1] = v2->y;
      coeff[0][2] = v2->z;

      coeff[9][0] = v3->x;
      coeff[9][1] = v3->y;
      coeff[9][2] = v3->z;

      /***********/
      /* Edge #1 */
      /***********/
      edge.x = v2->x - v1->x;
      edge.y = v2->y - v1->y;
      edge.z = v2->z - v1->z;

      //E - (E.N)N
      tangent1.x = edge.x;
      tangent1.y = edge.y;
      tangent1.z = edge.z;
      tmpf = (tangent1.x*n1->x) + (tangent1.y*n1->y) + (tangent1.z*n1->z);
      tangent1.x -= n1->x * tmpf;
      tangent1.y -= n1->y * tmpf;
      tangent1.z -= n1->z * tmpf;

      coeff[3][0] = v1->x + (tangent1.x * 0.3333f);
      coeff[3][1] = v1->y + (tangent1.y * 0.3333f);
      coeff[3][2] = v1->z + (tangent1.z * 0.3333f);

      //E - (E.N)N
      tangent2.x = edge.x;
      tangent2.y = edge.y;
      tangent2.z = edge.z;
      tmpf = (tangent2.x*n2->x) + (tangent2.y*n2->y) + (tangent2.z*n2->z);
      tangent2.x -= n2->x * tmpf;
      tangent2.y -= n2->y * tmpf;
      tangent2.z -= n2->z * tmpf;

      coeff[1][0] = v2->x - (tangent2.x * 0.3333f);
      coeff[1][1] = v2->y - (tangent2.y * 0.3333f);
      coeff[1][2] = v2->z - (tangent2.z * 0.3333f);

      /***********/
      /* Edge #2 */
      /***********/
      edge.x = v3->x - v2->x;
      edge.y = v3->y - v2->y;
      edge.z = v3->z - v2->z;

      //E - (E.N)N
      tangent1.x = edge.x;
      tangent1.y = edge.y;
      tangent1.z = edge.z;
      tmpf = (tangent1.x*n2->x) + (tangent1.y*n2->y) + (tangent1.z*n2->z);
      tangent1.x -= n2->x * tmpf;
      tangent1.y -= n2->y * tmpf;
      tangent1.z -= n2->z * tmpf;

      coeff[2][0] = v2->x + (tangent1.x * 0.3333f);
      coeff[2][1] = v2->y + (tangent1.y * 0.3333f);
      coeff[2][2] = v2->z + (tangent1.z * 0.3333f);

      //E - (E.N)N
      tangent2.x = edge.x;
      tangent2.y = edge.y;
      tangent2.z = edge.z;
      tmpf = (tangent2.x*n3->x) + (tangent2.y*n3->y) + (tangent2.z*n3->z);
      tangent2.x -= n3->x * tmpf;
      tangent2.y -= n3->y * tmpf;
      tangent2.z -= n3->z * tmpf;

      coeff[5][0] = v3->x - (tangent2.x * 0.3333f);
      coeff[5][1] = v3->y - (tangent2.y * 0.3333f);
      coeff[5][2] = v3->z - (tangent2.z * 0.3333f);

      /***********/
      /* Edge #3 */
      /***********/
      edge.x = v3->x - v1->x;
      edge.y = v3->y - v1->y;
      edge.z = v3->z - v1->z;

      //E - (E.N)N
      tangent1.x = edge.x;
      tangent1.y = edge.y;
      tangent1.z = edge.z;
      tmpf = (tangent1.x*n1->x) + (tangent1.y*n1->y) + (tangent1.z*n1->z);
      tangent1.x -= n1->x * tmpf;
      tangent1.y -= n1->y * tmpf;
      tangent1.z -= n1->z * tmpf;

      coeff[7][0] = v1->x + (tangent1.x * 0.3333f);
      coeff[7][1] = v1->y + (tangent1.y * 0.3333f);
      coeff[7][2] = v1->z + (tangent1.z * 0.3333f);

      //E - (E.N)N
      tangent2.x = edge.x;
      tangent2.y = edge.y;
      tangent2.z = edge.z;
      tmpf = (tangent2.x*n3->x) + (tangent2.y*n3->y) + (tangent2.z*n3->z);
      tangent2.x -= n3->x * tmpf;
      tangent2.y -= n3->y * tmpf;
      tangent2.z -= n3->z * tmpf;

      coeff[8][0] = v3->x - (tangent2.x * 0.3333f);
      coeff[8][1] = v3->y - (tangent2.y * 0.3333f);
      coeff[8][2] = v3->z - (tangent2.z * 0.3333f);

      /****************/
      /* Middle Point */
      /****************/
      // Farin's Method
      //    v2  0
      //       1 2
      //      3 4 5
      // v1  6 7 8 9  v3
      coeff[4][0] = (coeff[1][0] + coeff[3][0] + coeff[2][0] + coeff[5][0] + coeff[7][0] + coeff[8][0]) / 4.0f;
      coeff[4][1] = (coeff[1][1] + coeff[3][1] + coeff[2][1] + coeff[5][1] + coeff[7][1] + coeff[8][1]) / 4.0f;
      coeff[4][2] = (coeff[1][2] + coeff[3][2] + coeff[2][2] + coeff[5][2] + coeff[7][2] + coeff[8][2]) / 4.0f;

      coeff[4][0] -= (coeff[0][0] + coeff[6][0] + coeff[9][0]) / 6.0f;
      coeff[4][1] -= (coeff[0][1] + coeff[6][1] + coeff[9][1]) / 6.0f;
      coeff[4][2] -= (coeff[0][2] + coeff[6][2] + coeff[9][2]) / 6.0f;

      /***********************************************************/
      /* Compute control mesh for quadratic normal interpolation */
      /***********************************************************/
      //    v2  0
      //       1 2
      //   v1 3 4 5 v3

      //3 corner normals
      normCoeff[3][0] = n1->x;
      normCoeff[3][1] = n1->y;
      normCoeff[3][2] = n1->z;

      normCoeff[0][0] = n2->x;
      normCoeff[0][1] = n2->y;
      normCoeff[0][2] = n2->z;

      normCoeff[5][0] = n3->x;
      normCoeff[5][1] = n3->y;
      normCoeff[5][2] = n3->z;

      /*********************/
      /* Normal for edge 1 */
      /*********************/
      avgNorm.x = n1->x + n2->x;
      avgNorm.y = n1->y + n2->y;
      avgNorm.z = n1->z + n2->z;

      //N - 2*(E.N)E
      edge.x = v2->x - v1->x;
      edge.y = v2->y - v1->y;
      edge.z = v2->z - v1->z;
      tmpf = sqrt((edge.x*edge.x) + (edge.y*edge.y) + (edge.z*edge.z));
      if (tmpf > 0.0f)
      {
         edge.x /= tmpf;
         edge.y /= tmpf;
         edge.z /= tmpf;

         tmpf = 2.0f * ((edge.x*avgNorm.x) + (edge.y*avgNorm.y) + (edge.z*avgNorm.z)); //2*(E.N)

         avgNorm.x -= tmpf*edge.x;
         avgNorm.y -= tmpf*edge.y;
         avgNorm.z -= tmpf*edge.z;
         tmpf = 1.0f / sqrt((avgNorm.x*avgNorm.x) + (avgNorm.y*avgNorm.y) + (avgNorm.z*avgNorm.z));
         normCoeff[1][0] = avgNorm.x * tmpf;
         normCoeff[1][1] = avgNorm.y * tmpf;
         normCoeff[1][2] = avgNorm.z * tmpf;
      }
      else //Don't reflect normal about plane
      {
         tmpf = 1.0f / sqrt((avgNorm.x*avgNorm.x) + (avgNorm.y*avgNorm.y) + (avgNorm.z*avgNorm.z));
         normCoeff[1][0] = avgNorm.x * tmpf;
         normCoeff[1][1] = avgNorm.y * tmpf;
         normCoeff[1][2] = avgNorm.z * tmpf;
      }

      /*********************/
      /* Normal for edge 2 */
      /*********************/
      avgNorm.x = n3->x + n2->x;
      avgNorm.y = n3->y + n2->y;
      avgNorm.z = n3->z + n2->z;

      //N - 2*(E.N)E
      edge.x = v3->x - v2->x;
      edge.y = v3->y - v2->y;
      edge.z = v3->z - v2->z;
      tmpf = sqrt((edge.x*edge.x) + (edge.y*edge.y) + (edge.z*edge.z));
      if (tmpf > 0.0f)
      {
         edge.x /= tmpf;
         edge.y /= tmpf;
         edge.z /= tmpf;

         tmpf = 2.0f * ((edge.x*avgNorm.x) + (edge.y*avgNorm.y) + (edge.z*avgNorm.z)); //2*(E.N)

         avgNorm.x -= tmpf*edge.x;
         avgNorm.y -= tmpf*edge.y;
         avgNorm.z -= tmpf*edge.z;
         tmpf = 1.0f / sqrt((avgNorm.x*avgNorm.x) + (avgNorm.y*avgNorm.y) + (avgNorm.z*avgNorm.z));
         normCoeff[2][0] = avgNorm.x * tmpf;
         normCoeff[2][1] = avgNorm.y * tmpf;
         normCoeff[2][2] = avgNorm.z * tmpf;
      }
      else //Don't reflect normal about plane
      {
         tmpf = 1.0f / sqrt((avgNorm.x*avgNorm.x) + (avgNorm.y*avgNorm.y) + (avgNorm.z*avgNorm.z));
         normCoeff[2][0] = avgNorm.x * tmpf;
         normCoeff[2][1] = avgNorm.y * tmpf;
         normCoeff[2][2] = avgNorm.z * tmpf;
      }

      /*********************/
      /* Normal for edge 3 */
      /*********************/
      avgNorm.x = n1->x + n3->x;
      avgNorm.y = n1->y + n3->y;
      avgNorm.z = n1->z + n3->z;

      //N - 2*(E.N)E
      edge.x = v1->x - v3->x;
      edge.y = v1->y - v3->y;
      edge.z = v1->z - v3->z;
      tmpf = sqrt((edge.x*edge.x) + (edge.y*edge.y) + (edge.z*edge.z));
      if (tmpf > 0.0f)
      {
         edge.x /= tmpf;
         edge.y /= tmpf;
         edge.z /= tmpf;

         tmpf = 2.0f * ((edge.x*avgNorm.x) + (edge.y*avgNorm.y) + (edge.z*avgNorm.z)); //2*(E.N)

         avgNorm.x -= tmpf*edge.x;
         avgNorm.y -= tmpf*edge.y;
         avgNorm.z -= tmpf*edge.z;
         tmpf = 1.0f / sqrt((avgNorm.x*avgNorm.x) + (avgNorm.y*avgNorm.y) + (avgNorm.z*avgNorm.z));
         normCoeff[4][0] = avgNorm.x * tmpf;
         normCoeff[4][1] = avgNorm.y * tmpf;
         normCoeff[4][2] = avgNorm.z * tmpf;
      }
      else //Don't reflect normal about plane
      {
         tmpf = 1.0f / sqrt((avgNorm.x*avgNorm.x) + (avgNorm.y*avgNorm.y) + (avgNorm.z*avgNorm.z));
         normCoeff[4][0] = avgNorm.x * tmpf;
         normCoeff[4][1] = avgNorm.y * tmpf;
         normCoeff[4][2] = avgNorm.z * tmpf;
      }

      /************************/  //   b2
      /* Compute new vertices */  //
      /************************/  // b1  b3
      float32 b1 = 0.0f;
      float32 b2 = 1.0f;
      float32 b3 = 0.0f;
      float32 inc = 1.0f / (level+1);
      float32 numVerticesBeforeAdd = gNumRenderVertices; //Used below for indices
      for (j=0; j<(level+2); j++) //For each row of vertices
      {
         b1 = inc*j;
         b3 = 1.0f - b1 - b2;

         for (k=0; k<j+1; k++) //For each vertex in that row
         {
            /**********************/
            /* Compute new vertex */
            /**********************/
            EvaluateBezierPoint(coeff, normCoeff, b1, b2, b3, destVertices, destNormals);
            destVertices = (AtiVector3*)((uint32)destVertices + sizeof(NPatchVertex));

            if (gLinearNormals == TRUE) //Linear normals
            {
               destNormals->x = n1->x*b1 + n2->x*b2 + n3->x*b3;
               destNormals->y = n1->y*b1 + n2->y*b2 + n3->y*b3;
               destNormals->z = n1->z*b1 + n2->z*b2 + n3->z*b3;
               tmpf = 1.0f / sqrt(destNormals->x*destNormals->x + destNormals->y*destNormals->y + destNormals->z*destNormals->z);
               destNormals->x *= tmpf;
               destNormals->y *= tmpf;
               destNormals->z *= tmpf;
            }
            destNormals = (AtiVector3*)((uint32)destNormals + sizeof(NPatchVertex));

            destUVs->u = uv1->u*b1 + uv2->u*b2 + uv3->u*b3;
            destUVs->v = uv1->v*b1 + uv2->v*b2 + uv3->v*b3;
            destUVs = (AtiTexCoord2D*)((uint32)destUVs + sizeof(NPatchVertex));

            gNumRenderVertices++;

            //printf("%f %f %f\n", b1, b2, b3);
            b1 -= inc;
            b3 += inc;
         }

         b2 -= inc;
      }

      /************************/
      /* Generate new indices */
      /************************/
      uint16 x = numVerticesBeforeAdd + 0;
      uint16 y = numVerticesBeforeAdd + 1;
      for (j=0; j<(level+1); j++) //For each row of triangles
      {
         uint16 indices[3];

         indices[0] = y;
         indices[1] = x;

         /********************************/
         /* Add all polygons in this row */
         /********************************/
         for (k=0; k<2*j+1; k++)
         {
            //Get new index
            if (k & 0x1)
               indices[2] = ++x;
            else
               indices[2] = ++y;

            //printf("%d %d %d\n", indices[0], indices[1], indices[2]);

            //Add triangle
            destIndices[0] = indices[0];
            destIndices[1] = indices[1];
            destIndices[2] = indices[2];
            destIndices += 3;
            gNumRenderIndices += 3;

            //Swap indices for proper ordering
            if (k & 0x1)
               indices[1] = indices[2];
            else
               indices[0] = indices[2];
         }

         x++;
         y++;
      }
   }
}

//=============================================================================
void LoadModel (char *fileName)
{
   FILE *fp;
   char8 buff[128];
   int32 i;

   float32 minx = 9999999.0f;
   float32 maxx = -9999999.0f;
   float32 miny = 9999999.0f;
   float32 maxy = -9999999.0f;
   float32 minz = 9999999.0f;
   float32 maxz = -9999999.0f;

   /*************************/
   /* Open file for reading */
   /*************************/
   fp = fopen(fileName, "r");
   if (fp == NULL)
   {
      printf("Cannot open .m file for reading!\n");
      return;
   }

   /**********************/
   /* Save off file name */
   /**********************/
   strcpy(gCurrentFile, fileName);

   /***************/
   /* Free memory */
   /***************/
   AtiFreeTypeArray(gVertexArray);
   gNumVertices = 0;
   gNumVerticesAllocated = 0;

   AtiFreeTypeArray(gIndexArray);
   gNumIndices = 0;
   gNumIndicesAllocated = 0;

   /***********************/
   /* Read in entire file */
   /***********************/
   for (;;)
   {
      /*****************************/
      /* Allocate memory if needed */
      /*****************************/
      if (gNumVertices >= gNumVerticesAllocated)
      {
         NPatchVertex *tmpArray = gVertexArray;

         gNumVerticesAllocated += 50;
         AtiMallocTypeArray (gVertexArray, gNumVerticesAllocated, NPatchVertex);
         memcpy (gVertexArray, tmpArray, sizeof(NPatchVertex)*gNumVertices);
         AtiFreeTypeArray (tmpArray);
      }

      if (gNumIndices >= gNumIndicesAllocated)
      {
         uint16 *tmpArray = gIndexArray;

         gNumIndicesAllocated += 50;
         AtiMallocTypeArray (gIndexArray, gNumIndicesAllocated, uint16);
         memcpy (gIndexArray, tmpArray, sizeof(uint16)*gNumIndices);
         AtiFreeTypeArray (tmpArray);
      }

      /*************************/
      /* Read in the next line */
      /*************************/
      if (fgets(buff, 128, fp) != NULL) //Read in line from file
      {
         if (buff[0] == 'V')
         {
            NPatchVertex *vert = &(gVertexArray[gNumVertices]);
            
            sscanf(buff, "V %f %f %f %f %f %f %f %f",
               &(vert->vertex.x), &(vert->vertex.y), &(vert->vertex.z),
               &(vert->normal.x), &(vert->normal.y), &(vert->normal.z),
               &(vert->uv.u), &(vert->uv.v));

            if (vert->vertex.x < minx)
               minx = vert->vertex.x;
            if (vert->vertex.x > maxx)
               maxx = vert->vertex.x;

            if (vert->vertex.y < miny)
               miny = vert->vertex.y;
            if (vert->vertex.y > maxy)
               maxy = vert->vertex.y;

            if (vert->vertex.z < minz)
               minz = vert->vertex.z;
            if (vert->vertex.z > maxz)
               maxz = vert->vertex.z;

            gNumVertices++;
         }
         else if (buff[0] == 'I')
         {
            int32 tmpi;
            sscanf(buff, "I %d", &tmpi);
            gIndexArray[gNumIndices] = (uint16)tmpi;
            gNumIndices++;
         }
         else //If end of file
         {
            break;
         }
      }
      else
      {
         break;
      }
   }

   /****************************/
   /* Scale and bias the model */
   /****************************/
   float32 min;
   float32 max;
   if (minx < miny)
      min = minx;
   else
      min = miny;
   if (minz < min)
      min = minz;

   if (maxx > maxy)
      max = maxx;
   else
      max = maxy;
   if (maxz > max)
      max = maxz;

   float32 scalex = 10.0f / (maxx - minx);
   float32 scaley = 10.0f / (maxy - miny);
   float32 scalez = 10.0f / (maxz - minz);
   float32 scale = (scalex < scaley) ? scalex : scaley;
   scale = (scalez < scale) ? scalez : scale;
   float32 biasx = -(minx + ((maxx - minx)*0.5f));
   float32 biasy = -(miny + ((maxy - miny)*0.5f));
   float32 biasz = -(minz + ((maxz - minz)*0.5f));

   for (i=0; i<gNumVertices; i++)
   {
      gVertexArray[i].vertex.x = (gVertexArray[i].vertex.x + biasx) * scale;
      gVertexArray[i].vertex.y = (gVertexArray[i].vertex.y + biasy) * scale;
      gVertexArray[i].vertex.z = (gVertexArray[i].vertex.z + biasz) * scale;
   }

   /******************/
   /* N-Patch object */
   /******************/
   CreateNPatchGeometry(gLevel);
}

//=============================================================================
void Draw (void)
{
   uint32 i;
   uint32 j;
   float32 matrix[16];

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glTranslatef(gPanX, gPanY, -gZoomDistance);
   glRotatef(gPitch, 1.0f, 0.0f, 0.0f);
   glRotatef(gYaw, 0.0f, 1.0f, 0.0f);
   glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);

   //Get camera position
   glGetFloatv(GL_MODELVIEW_MATRIX, matrix);
   gCamera.x = 3.0f * matrix[2];
   gCamera.y = 3.0f * matrix[6];
   gCamera.z = 3.0f * matrix[10];

   /***********************/
   /* Draw original model */
   /***********************/
   glViewport(0, 0, gWindowWidth/2, gWindowHeight);

   glEnableClientState(GL_VERTEX_ARRAY);
   glVertexPointer(3, GL_FLOAT, sizeof(NPatchVertex), &(gVertexArray[0].vertex));

   glEnableClientState(GL_NORMAL_ARRAY);
   glNormalPointer(GL_FLOAT, sizeof(NPatchVertex), &(gVertexArray[0].normal));

   glDrawElements(GL_TRIANGLES, gNumIndices, GL_UNSIGNED_SHORT, gIndexArray);

   if (gDrawWireframe == TRUE)
   {
      glColor3f(0.0f, 0.0f, 0.0f);
      glDisable(GL_LIGHTING);
      glPolygonMode(GL_FRONT, GL_LINE);
      glPolygonOffset(-1.0f, 0.0f);
      glEnable(GL_POLYGON_OFFSET_LINE);
      glDrawElements(GL_TRIANGLES, gNumIndices, GL_UNSIGNED_SHORT, gIndexArray);
      glDisable(GL_POLYGON_OFFSET_LINE);
      glPolygonMode(GL_FRONT, GL_FILL);
      glEnable(GL_LIGHTING);
   }

   glDisableClientState(GL_VERTEX_ARRAY);
   glDisableClientState(GL_NORMAL_ARRAY);

   if (gDrawNormals == TRUE)
   {
      glColor3f(1.0f, 0.0f, 0.0f);
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);
      {
         NPatchVertex *vert = gVertexArray;
         for (i=0; i<(uint32)gNumVertices; i++)
         {
            glVertex3f(vert->vertex.x, vert->vertex.y, vert->vertex.z);
            glVertex3f(vert->vertex.x+vert->normal.x*0.3f, vert->vertex.y+vert->normal.y*0.3f, vert->vertex.z+vert->normal.z*0.3f);
            vert++;
         }
      }
      glEnd();
      glEnable(GL_LIGHTING);
   }

   /************************/
   /* Draw N-Patched model */
   /************************/
   glViewport(gWindowWidth/2, 0, gWindowWidth/2, gWindowHeight);

   glEnableClientState(GL_VERTEX_ARRAY);
   glVertexPointer(3, GL_FLOAT, sizeof(NPatchVertex), &(gRenderVertexArray[0].vertex));

   glEnableClientState(GL_NORMAL_ARRAY);
   glNormalPointer(GL_FLOAT, sizeof(NPatchVertex), &(gRenderVertexArray[0].normal));

   glDrawElements(GL_TRIANGLES, gNumRenderIndices, GL_UNSIGNED_SHORT, gRenderIndexArray);

   if (gDrawWireframe == TRUE)
   {
      glColor3f(0.0f, 0.0f, 0.0f);
      glDisable(GL_LIGHTING);
      glPolygonMode(GL_FRONT, GL_LINE);
      glPolygonOffset(-1.0f, 0.0f);
      glEnable(GL_POLYGON_OFFSET_LINE);
      glDrawElements(GL_TRIANGLES, gNumRenderIndices, GL_UNSIGNED_SHORT, gRenderIndexArray);
      glDisable(GL_POLYGON_OFFSET_LINE);
      glPolygonMode(GL_FRONT, GL_FILL);
      glEnable(GL_LIGHTING);
   }

   glDisableClientState(GL_VERTEX_ARRAY);
   glDisableClientState(GL_NORMAL_ARRAY);

   if (gDrawNormals == TRUE)
   {
      glColor3f(1.0f, 0.0f, 0.0f);
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);
      {
         NPatchVertex *vert = gRenderVertexArray;
         for (i=0; i<(uint32)gNumRenderVertices; i++)
         {
            glVertex3f(vert->vertex.x, vert->vertex.y, vert->vertex.z);
            glVertex3f(vert->vertex.x+vert->normal.x*0.3f, vert->vertex.y+vert->normal.y*0.3f, vert->vertex.z+vert->normal.z*0.3f);
            vert++;
         }
      }
      glEnd();
      glEnable(GL_LIGHTING);
   }

   /******************/
   /* Reset Viewport */
   /******************/
   glViewport(0, 0, gWindowWidth, gWindowHeight);

   /********************************/
   /* Draw separator line and text */
   /********************************/
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   gluOrtho2D(0.0, 100.0, 0.0, 100.0);

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();
   glLineWidth(2.0f);

   glDisable(GL_LIGHTING);

   glColor3f(0.0f, 0.0f, 0.0f);
   glBegin(GL_LINES);
     glVertex2f(50.0f, 0.0f);
     glVertex2f(50.0f, 100.0f);
   glEnd();

   //Draw text
   char8 buff[64];

   sprintf(buff, "Input Model (%d polygons)", gNumIndices/3);
   glLoadIdentity();
   glTranslatef(strlen(buff)*0.27, 1.0f, 0.0f);
   glScalef(0.010f, 0.015f, 1.0f);
   for (i=0; i<strlen(buff); i++)
      glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, buff[i]);

   sprintf(buff, "N-Patch Level %d (%d polygons)", gLevel, gNumRenderIndices/3);
   glLoadIdentity();
   glTranslatef(50.0f+(strlen(buff)*0.27), 1.0f, 0.0f);
   glScalef(0.010f, 0.015f, 1.0f);
   for (i=0; i<strlen(buff); i++)
      glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, buff[i]);

   if (gLinearNormals == TRUE)
      sprintf(buff, "Linear Normals (Press L)");
   else
      sprintf(buff, "Quadratic Normals (Press L)");
   glLoadIdentity();
   glTranslatef(63.0f, 4.0f, 0.0f);
   glScalef(0.010f, 0.015f, 1.0f);
   for (i=0; i<strlen(buff); i++)
      glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, buff[i]);

   if (gAutoLODMode == TRUE)
      sprintf(buff, "Auto LOD");
   else
      sprintf(buff, "Manual LOD");
   glLoadIdentity();
   glTranslatef(71.0f, 98.0f, 0.0f);
   glScalef(0.010f, 0.015f, 1.0f);
   for (i=0; i<strlen(buff); i++)
      glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, buff[i]);

   if (gShowHelp == TRUE)
   {
      float32 y = 98.0f;
      for (i=0; i<11; i++)
      {
         switch (i)
         {
            case 0: sprintf(buff, "0-9 - Load Models"); break;
            case 1: sprintf(buff, "A - Toggle Auto LOD (Based on Distance)"); break;
            case 2: sprintf(buff, "N - Toggle Normals"); break;
            case 3: sprintf(buff, "W - Toggle Wireframe"); break;
            case 4: sprintf(buff, "S - Toggle Specular"); break;
            case 5: sprintf(buff, "L - Toggle Linear/Quadratic Normal Interpolation"); break;
            case 6: sprintf(buff, "H - Show/Hide Help"); break;
            case 7: sprintf(buff, "Left Mouse Button - Hold & Drag to Rotate"); y -= 2.0f; break;
            case 8: sprintf(buff, "Right Mouse Button - Hold & Drag to Zoom"); break;
            case 9: sprintf(buff, "Both Mouse Buttons - Hold & Drag to Pan"); break;
            case 10: sprintf(buff, "Up/Down Keys - Increase/Decrease Tessellation (Manual LOD)"); y -= 2.0f; break;
         }

         glLoadIdentity();
         glTranslatef(1.0f, y, 0.0f);
         glScalef(0.012f, 0.015f, 1.0f);
         for (j=0; j<strlen(buff); j++)
            glutStrokeCharacter(GLUT_STROKE_ROMAN, buff[j]);
         y -= 2.0f;
      }
   }
   else
   {
      sprintf(buff, "Press 'H' for help");
      glLoadIdentity();
      glTranslatef(1.0f, 98.0f, 0.0f);
      glScalef(0.010f, 0.010f, 1.0f);
      glLineWidth(1.0f);
      for (i=0; i<strlen(buff); i++)
         glutStrokeCharacter(GLUT_STROKE_ROMAN, buff[i]);
   }

   //Reset state
   glEnable(GL_LIGHTING);

   glLineWidth(1.0f);
   glPopMatrix();
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);

   /****************/
   /* Swap buffers */
   /****************/
   glutSwapBuffers();
}

//=============================================================================
void SetLights (void)
{
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();

   //Lights and materials
   float32 gLight0Position[] = {-40.0f, 10.0f,  80.0f, 1.0f};
   float32 gLight0Ambient[]  = {0.8f, 0.8f, 0.8f, 1.0f};
   float32 gLight0Diffuse[]  = {1.0f, 1.0f, 1.0f, 1.0f};
   float32 gLight0Specular[] = {1.0f, 1.0f, 1.0f, 1.0f};

   float32 gLight1Position[] = { 90.5f, 70.0f,-100.5f, 1.0f};
   float32 gLight1Ambient[]  = {0.3f, 0.1f, 0.1f, 1.0f};
   float32 gLight1Diffuse[]  = {0.3f, 0.1f, 0.1f, 1.0f};
   float32 gLight1Specular[] = {1.0f, 0.3f, 0.3f, 1.0f};

   glLightfv(GL_LIGHT0, GL_POSITION, gLight0Position);
   glLightfv(GL_LIGHT1, GL_POSITION, gLight1Position);

   glLightfv(GL_LIGHT0, GL_AMBIENT,  gLight0Ambient);
   glLightfv(GL_LIGHT0, GL_DIFFUSE,  gLight0Diffuse);
   glLightfv(GL_LIGHT0, GL_SPECULAR, gLight0Specular);

   glLightfv(GL_LIGHT1, GL_AMBIENT,  gLight1Ambient);
   glLightfv(GL_LIGHT1, GL_DIFFUSE,  gLight1Diffuse);
   glLightfv(GL_LIGHT1, GL_SPECULAR, gLight1Specular);

   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   glEnable(GL_LIGHT1);
   glShadeModel(GL_SMOOTH);

   float32 ambient[] = {0.1f, 0.1f, 0.1f, 1.0f};
   float32 diffuse[] = {0.6f, 0.6f, 0.6f, 1.0f};
   float32 specularOn[] = {1.0f, 1.0f, 1.0f, 1.0f};
   float32 specularOff[] = {0.0f, 0.0f, 0.0f, 0.0f};
   glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);
   glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
   if (gDrawSpecular == TRUE)
      glMaterialfv(GL_FRONT, GL_SPECULAR, specularOn);
   else
      glMaterialfv(GL_FRONT, GL_SPECULAR, specularOff);
   glMaterialf(GL_FRONT, GL_SHININESS, 80.0f);

   glPopMatrix();
}

//=============================================================================
int32 MyInit (void)
{
   //Set up OpenGL states
   glEnable(GL_DEPTH_TEST); //Turn on z-buffer
   glFrontFace(GL_CCW);     //Counter Clock-wise polygons face out
   glEnable(GL_CULL_FACE);  //Cull faces based on glFrontFace
   glCullFace(GL_BACK);
   glDepthFunc(GL_LEQUAL);
   glDisable(GL_LINE_SMOOTH);
   glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
   glDrawBuffer(GL_BACK);
   glClearColor(0.9f, 0.9f, 0.9f, 0.0f);
   glDisable(GL_TEXTURE_2D);
   glPolygonMode(GL_FRONT, GL_FILL);
   glReadBuffer(GL_BACK);

   //Lights
   SetLights();

   //Camera
   gCamera.x = 0.0f;
   gCamera.y = 0.0f;
   gCamera.z = 0.0f;

   //Load first model
   LoadModel("1.m");

   return TRUE;
}

//=============================================================================
void ComputeAndSetAutoLOD (void)
{
   float32 newLevel = (500.0f-gZoomDistance)/500.0f;
   newLevel = newLevel*newLevel*newLevel*10.0f;
   if (newLevel < 1.0f) //Slow down the 1 to 0 level since the poly difference is more drastic
      newLevel *= 2.0f;
   if ((int32)newLevel != gLevel)
      CreateNPatchGeometry(newLevel);
}

//=============================================================================
void MouseFunc(int32 button, int32 state, int32 x, int32 y)
{
   static lButtonDown = FALSE;
   static rButtonDown = FALSE;
   static lastX = 0;
   static lastY = 0;

   switch(button)
   {
      case GLUT_LEFT_BUTTON:
         if (state == GLUT_DOWN)
         {
            lastX = x;
            lastY = y;
            lButtonDown = TRUE;
         }
         else
            lButtonDown = FALSE;
         break;

      case GLUT_RIGHT_BUTTON:
         if (state == GLUT_DOWN)
         {
            lastX = x;
            lastY = y;
            rButtonDown = TRUE;
         }
         else
            rButtonDown = FALSE;
         break;
   }


   if ((lButtonDown == TRUE) && (rButtonDown == TRUE))
   {
      gPanX -= (float)(lastX - x) / (float)gWindowWidth * 10.0f;
      gPanY += (float)(lastY - y) / (float)gWindowHeight * 10.0f;

      lastX = x;
      lastY = y;
   }
   else if (lButtonDown == TRUE)
   {
      gYaw -= (float)(lastX - x) / (float)gWindowWidth * 400.0f;
      gPitch -= (float)(lastY - y) / (float)gWindowHeight * 200.0f;

      if (gYaw >= 360.0f)
         gYaw -= 360.0f;
      if (gYaw <= 0.0f)
         gYaw += 360.0f;

      if (gPitch < -90.0f)
         gPitch = -90.0f;
      if (gPitch > 90.0f)
         gPitch = 90.0f;

      lastX = x;
      lastY = y;
   }
   else if (rButtonDown == TRUE)
   {
      gZoomDistance += (float)(lastY - y) / (float)gWindowHeight * 300.0f;

      if (gZoomDistance < 1.0f)
         gZoomDistance = 1.0f;
      if (gZoomDistance > 900.0f)
         gZoomDistance = 900.0f;

      if (gAutoLODMode == TRUE)
      {
         ComputeAndSetAutoLOD();
      }

      lastX = x;
      lastY = y;
   }
}

//=============================================================================
void MotionFunc(int32 x, int32 y)
{
   MouseFunc(999, 0, x, y);
}

//=============================================================================
// Key input handler
//=============================================================================
void KeyFunc (unsigned char key, int32 x, int32 y)
{
   switch(key)
   {
      //Quit
      case '\x1b':
      case 'Q':
      case 'q':
         exit(0);
         break;

      case 'a':
      case 'A':
         gAutoLODMode ^= TRUE;
         if (gAutoLODMode == TRUE)
            ComputeAndSetAutoLOD();
         break;

      case 'h':
      case 'H':
         gShowHelp ^= TRUE;
         break;

      case 'l':
      case 'L':
         gLinearNormals ^= TRUE;
         CreateNPatchGeometry(gLevel);
         break;

      case 'n':
      case 'N':
         gDrawNormals ^= TRUE;
         break;

      case 's':
      case 'S':
         gDrawSpecular ^= TRUE;
         SetLights();
         break;

      case 'w':
      case 'W':
         gDrawWireframe ^= TRUE;
         break;

      case '1': LoadModel("1.m"); break;
      case '2': LoadModel("2.m"); break;
      case '3': LoadModel("3.m"); break;
      case '4': LoadModel("4.m"); break;
      case '5': LoadModel("5.m"); break;
      case '6': LoadModel("6.m"); break;
      case '7': LoadModel("7.m"); break;
      case '8': LoadModel("8.m"); break;
      case '9': LoadModel("9.m"); break;
      case '0': LoadModel("0.m"); break;
   }
}

//=============================================================================
void SpecialFunc (int32 key, int32 x, int32 y)
{
   switch(key)
   {
      case GLUT_KEY_UP:
         if (gAutoLODMode == FALSE)
            CreateNPatchGeometry(gLevel+1);
         break;

      case GLUT_KEY_DOWN:
         if (gAutoLODMode == FALSE)
            CreateNPatchGeometry(gLevel-1);
         break;
   }
}

//=============================================================================
// Projection and View Setup
//=============================================================================
void MyReshape(int32 w, int32 h)
{
   gWindowWidth = w;
   gWindowHeight = h;

   glViewport(0,0,w,h);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(40.0, 0.5f*(double)w/(double)h, 1.0, 10000.0);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glTranslatef(0.0f,0.0f,-5.0f);
}

//=============================================================================
int32 main(int32 argc, char **argv)
{
   printf("ATI N-Patch Viewer\n");
   printf("March, 2001\n");
   printf("Author: Alex Vlachos (AVlachos@ati.com)\n");
   printf("\n");
   printf(" ESC      - Quit program\n");
   printf("\n");
   printf(" 0-9 - Load Models\n");
   printf("\n");
   printf(" A - Toggle Auto LOD Based on Distance (Default Off)\n");
   printf(" N - Toggle Normals (Default Off)\n");
   printf(" W - Toggle Wireframe (Default Off)\n");
   printf(" S - Toggle Specular (Default On)\n");
   printf(" L - Toggle Linear/Quadratic Normal Interpolation\n");
   printf("\n");
   printf(" Up Arrow - Increase Tessellation (Stops when model reaches 65k vertices)\n");
   printf(" Down Arrow - Decrease Tessellation (Stops at level 0)\n");

   GetWindowRect(GetDesktopWindow(), &gDesktopRect);

   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
   glutCreateWindow("ATI N-Patch Viewer");
   glutReshapeWindow(gDesktopRect.right-8, gDesktopRect.bottom); //gWindowWidth, gWindowHeight);
   glutDisplayFunc(Draw);
   glutReshapeFunc(MyReshape);
   glutKeyboardFunc(KeyFunc);
   glutSpecialFunc(SpecialFunc);
   glutMouseFunc(MouseFunc);
   glutMotionFunc(MotionFunc);
   glutIdleFunc(Draw);
   glutFullScreen();

   if (MyInit() == FALSE)
      return 0;

   glutMainLoop();

   /***************/
   /* Free memory */
   /***************/
   AtiFreeTypeArray(gVertexArray);
   AtiFreeTypeArray(gIndexArray);
   AtiFreeTypeArray(gRenderVertexArray);
   AtiFreeTypeArray(gRenderIndexArray);

   return 0;
}
