/******************************************************************************
 * AtiMAXVertexNormals.cpp
 * NOTE: All code taken from the MAX 3.1 SDK (give or take some formatting).
 *    The only change was a fix when no smoothing groups are specified.
 ******************************************************************************
 $Header:: /N-Patch/Max3.1Exporter/AtiMAXVertexNormals.cpp 2     4/13/01 12:3 $
 ******************************************************************************
 * Created by Alex Vlachos (AVlachos@ati.com) on 4-12-2001
 * (C) 2000 ATI Research, Inc.  All rights reserved.
 ******************************************************************************/

// INCLUDES ===================================================================
#include "AtiMAXVertexNormals.h"

//=============================================================================
// Add a normal to the list if the smoothing group bits overlap, 
// otherwise create a new vertex normal in the list
//=============================================================================
void VNormal::AddNormal(Point3 &n, DWORD s)
{
   if (((s & smooth) == 0) && (init)) //If different smGroup and already init'd
   {
      if (next)
         next->AddNormal(n,s);
      else
         next = new VNormal(n,s);
   }
   else
   {
      norm   += n;
      smooth |= s;
      init    = TRUE;
   }
}

//=============================================================================
// Retrieves a normal if the smoothing groups overlap or there is
// only one in the list
//=============================================================================
Point3 &VNormal::GetNormal(DWORD s)
{
   if ((smooth & s) || (!next))
      return norm;
   else
      return next->GetNormal(s);
}

//=============================================================================
// Normalize each normal in the list
//=============================================================================
void VNormal::Normalize()
{
   VNormal *ptr = next, *prev = this;
   while (ptr)
   {
      if (ptr->smooth & smooth)
      {
         norm += ptr->norm;
         prev->next = ptr->next;
         delete ptr;
         ptr = prev->next;
      } 
      else
      {
         prev = ptr;
         ptr  = ptr->next;
      }
   }
   norm = ::Normalize(norm);
   if (next)
      next->Normalize();
}

//=============================================================================
// Corner is 0, 1, or 2 -- which corner do we want the angle of?
//=============================================================================
float AtiMAXFindVertexAngle(Mesh *mesh, int face, int corner)
{
   int cnext = (corner+1)%3;
   int cprev = (corner+2)%3;
   DWORD *vv = mesh->faces[face].v;

   // Get edge vectors:
   Point3 A = mesh->verts[vv[cnext]] - mesh->verts[vv[corner]];
   Point3 B = mesh->verts[vv[corner]] - mesh->verts[vv[cprev]];

   // Normalize the edge-vectors, but return 0 if either has 0 length.
   float len = Length(A);
   if (!len)
      return len;
   A = A/len;

   len = Length(B);
   if (!len)
      return len;
   B = B/len;

   // The dot product gives the cosine of the angle:
   float dp = DotProd (A,B);
   if (dp>1) dp=1.0f;   // shouldn't happen, but might
   if (dp<-1) dp=-1.0f;   // shouldn't happen, but might
   return acos(dp);
}
