/******************************************************************************
 * SushiExp.h
 ******************************************************************************
 $Header:: /N-Patch/Max3.1Exporter/NPatchExp.h 3     4/13/01 12:34a Avlachos  $
 ******************************************************************************
 * Created by Alex Vlachos (AVlachos@ati.com) on 4-12-2001
 * (C) 2000 ATI Research, Inc.  All rights reserved.
 ******************************************************************************/

#ifndef __SUSHIEXP__H
#define __SUSHIEXP__H

// INCLUDES ===================================================================
#include "Max.h"
#include "msv/resource.h"
#include "istdplug.h"
#include "stdmat.h"
#include "shape.h"
#include "Types.h"

// DEFINES ====================================================================

// FUNCTION DEFS ==============================================================
extern ClassDesc* GetSushiOutDesc();
extern TCHAR *GetString(int id);
extern HINSTANCE hInstance;

// STRUCTURES =================================================================
typedef union
{
   //Point in 3D space
   struct { float32 x, y, z; };
   struct { float32 v[3]; };
} AtiVector3;

typedef union
{
   //Point in 2D space
   struct { float32 u, v; };
   struct { float32 t[2]; };
} AtiTexCoord2D;

typedef struct //Preprocess polygon
{
   AtiVector3 vertex[3];
   AtiVector3 normal[3];
   AtiTexCoord2D uv[3];
} NPatchPolygon;

typedef struct //Preprocess polygon
{
   AtiVector3 vertex;
   AtiVector3 normal;
   AtiTexCoord2D uv;
} NPatchVertex;

// MACROS =====================================================================
#define AtiMallocTypeArray(array,num,type)                               \
   array = NULL;                                                         \
   if ((num) > 0)                                                        \
   {                                                                     \
      array = new type [(num)];                                            \
      /*array = (type *) malloc (sizeof(type) * (num));*/                \
      if (array == NULL)                                                 \
      {                                                                  \
         SuErrorLog("\nERROR: Can't allocate memory (%d bytes) for " #array "!\n", sizeof(type) * (num)); \
         return;                                                         \
      }                                                                  \
      memset (array, 0, sizeof(type) * (num));                           \
   }

#define AtiFreeTypeArray(array) \
   if (array != NULL)           \
      /*free(array);*/          \
      delete [] array;          \
   array = NULL;

// CLASS DEFINITION ===========================================================
class SushiOut : public SceneExport
{
   public:
      SushiOut();
      ~SushiOut();

      Interface* ip;

      int32 mCurNode;
      int32 mTotalNodeCount;

      //N-Patch variables
      MtlList mtlList;
      Color mLightTintColor;
      int32 mNumLights;
      
      //Brute-force organization
      int32 mNumPolygons;
      NPatchPolygon *mPolygonArray;

      //Arrays that are written to disk
      int32 mNumVertices;
      NPatchVertex *mVertexArray;

      int32 mNumIndices;
      uint32 *mIndexArray;

      void AddVertexToArray (NPatchPolygon *poly, int32 idx); //NOTE: Also adds the index to the index array

      //SceneExport methods
      int   ExtCount(); //Number of extensions supported 
      const TCHAR * Ext(int n); //Extension #n (i.e. "SU")
      const TCHAR * LongDesc(); //Long description (i.e. "Sushi Export") 
      const TCHAR * ShortDesc(); //Short description (i.e. "SushiExp")
      const TCHAR * AuthorName(); //Author name
      const TCHAR * CopyrightMessage(); //Copyright message 
      const TCHAR * OtherMessage1(); //Other message #1
      const TCHAR * OtherMessage2(); //Other message #2
      unsigned int Version(); //Version number * 100 (i.e. v3.01 = 301) 
      void  ShowAbout(HWND hWnd); //Show DLL's "About..." box
      int   DoExport(const TCHAR *name,ExpInterface *ei,Interface *i, BOOL suppressPrompts=FALSE, DWORD options=0); //Export file
      BOOL  SupportsOptions(int ext, DWORD options);

      //Node enumeration
      BOOL NodeEnum(INode* node);
      BOOL CountAllByNodeEnum(INode* node);

      //High level export
      void CountGeomObject(INode* node); 
      void CountLightObject(INode* node); 
      void CountHelperObject(INode* node);

      void ExportGlobalInfo();
      void ExportGeomObject(INode* node); 
      void ExportLightObject(INode* node); 
      void ExportCameraObject(INode* node); 
      void ExportShapeObject(INode* node); 
      void ExportHelperObject(INode* node);
   
      //Mid level export
      void ExportMesh(INode* node, TimeValue t); 
      void ExportNodeHeader(INode* node, TCHAR* type);
      void ExportCameraSettings(CameraState* cs, CameraObject* cam, TimeValue t);
      void ExportLightSettings(LightState* ls, GenLight* light, TimeValue t);
   
      //Misc methods
      TriObject* GetTriObjectFromNode(INode *node, TimeValue t, int &deleteIt);
   
      //Interface to member variables
      inline Interface* GetInterface() { return ip; }
};

#endif //__SUSHIEXP__H
